#ifndef _C4_COMMON_HPP_
#define _C4_COMMON_HPP_

#include "c4/config.hpp"

#include "c4/preprocessor.hpp"
#include "c4/platform.hpp"
#include "c4/cpu.hpp"
#include "c4/compiler.hpp"
#include "c4/language.hpp"
#include "c4/error.hpp"
#include "c4/time.hpp"
#include "c4/types.hpp"

#endif /* _C4_COMMON_HPP_ */
