// This is an open source non-commercial project. Dear PVS-Studio, please check it.
// PVS-Studio Static Code Analyzer for C, C++ and C#: http://www.viva64.com

#include "c4/libtest/archetypes.hpp"

C4_BEGIN_NAMESPACE(c4)
C4_BEGIN_NAMESPACE(archetypes)

int IdOwner::s_current = 0;

C4_END_NAMESPACE(archetypes)
C4_END_NAMESPACE(c4)
