// This is an open source non-commercial project. Dear PVS-Studio, please check it.
// PVS-Studio Static Code Analyzer for C, C++ and C#: http://www.viva64.com

#include "c4/test.hpp"

C4_BEGIN_NAMESPACE(c4)

size_t TestErrorOccurs::num_errors = 0;

C4_END_NAMESPACE(c4)
