#ifndef _C4_STD_STD_HPP_
#define _C4_STD_STD_HPP_

/** @file std.hpp includes all c4-std interop files */

#include "c4/std/vector.hpp"
#include "c4/std/string.hpp"
#include "c4/std/tuple.hpp"

#endif // _C4_STD_STD_HPP_
